﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

using System.Threading;
using System.Net.Sockets;
using System.Net;

using System.Diagnostics;



namespace IPRouter
{
    public partial class Form1 : Form
    {
        
        //Set some global variables
        string programVersion = "0.07 Beta";
        
        int internalPortBaseNET = 53248;        //internal port for NET device
        int internalPortBaseNx = 53504;         //internal port for Nx device
        int externalPortBaseNET = 7001;         //external port for NET device
        int externalPortBaseNx = 7002;          //external port for Nx device

        //string IPaddressAny = "127.0.0.1";          // Local Host
        string IPaddressAny = "0.0.0.0";            //IP address any
        string IPaddressLocal = "127.0.0.1";        //IP address of local host

        Thread[] jobArray = new Thread[4];          //holds the four main jobs
        Socket[] socketArray = new Socket[4];       //holds the sockets for the four main jobs

        //Un-comment for testing
        //string testExt = "12";

        public Form1()
        {
            InitializeComponent();
        }

        

        private void Form1_Load(object sender, EventArgs e)
        {
            // Set some default testing values
            txtinternalStnno.Text = "7";
            txtExternalIPAddress.Text = "172.16.0.2";
            
            Text = "IP Router V" + programVersion;

            btnStop.Enabled = false;
            btnStop.Visible = false;
            btnStart.Visible = true;
   
        }


        private void btnStart_Click(object sender, EventArgs e)
        {
            //Check that the internal network station number and external IP Address are in range
            if (!CheckNETStation(txtinternalStnno.Text))
            {
                return;
            } //end if

            if (!CheckIPAddress(txtExternalIPAddress.Text))
            {
                return;
            }
            
            //Set button and text box status
            btnStart.Enabled = false;
            btnStart.Visible = false;
            btnStop.Enabled = true;
            btnStop.Visible = true;

            txtinternalStnno.Enabled = false;
            txtExternalIPAddress.Enabled = false;

            //Set Port numbers for internal and external network types
            int iPortNET = internalPortBaseNET + int.Parse(txtinternalStnno.Text);
            int iPortNx = internalPortBaseNx + int.Parse(txtinternalStnno.Text); 

            int ePortNET = externalPortBaseNET;
            int ePortNx = externalPortBaseNx;

            // Un-comment for testing 
            //int ePortNET = internalPortBaseNET + int.Parse(testExt);
            //int ePortNx = internalPortBaseNx + int.Parse(testExt);
            

            //Open four sockets for the main jobs
            socketArray[0] = new Socket(AddressFamily
               .InterNetwork, SocketType.Stream, ProtocolType.Tcp);

            socketArray[1] = new Socket(AddressFamily
               .InterNetwork, SocketType.Stream, ProtocolType.Tcp);

            socketArray[2] = new Socket(AddressFamily
               .InterNetwork, SocketType.Stream, ProtocolType.Tcp);

            socketArray[3] = new Socket(AddressFamily
               .InterNetwork, SocketType.Stream, ProtocolType.Tcp);

            
            //Start four threads running to monitor for and accept connection requests
            jobArray[0] = new Thread(new ThreadStart(() => iNETTask(IPaddressLocal,iPortNET,0)));
            jobArray[0].Name = "INET";
            jobArray[0].Start();

            jobArray[1] = new Thread(new ThreadStart(() => iNxTask(IPaddressLocal, iPortNx, 0)));
            jobArray[1].Name = "INx";
            jobArray[1].Start();

            jobArray[2] = new Thread(new ThreadStart(() => eNETTask(IPaddressAny, ePortNET, 1)));
            jobArray[2].Name = "ENET";
            jobArray[2].Start();

            jobArray[3] = new Thread(new ThreadStart(() => eNxTask(IPaddressAny, ePortNx, 1)));
            jobArray[3].Name = "ENx";
            jobArray[3].Start();

        }


        private bool CheckNETStation(string x)
        {
            //Check the internal network station number
            int station;
            try
            {
                station = int.Parse(x);
            }
            catch (FormatException f)
            {
                // Deal with errors
                station = 0;
                if (f.Source != null)
                    Debug.WriteLine("FormatException source: {0}", f.Source);
            }

            if ((station < 1) || (station > 254))
            {
                //Station number is out of range
                string message = "Invalid network station number entered\n\r Must be between 1 and 254";
                MessageBoxButtons buttons = MessageBoxButtons.OK;
                MessageBox.Show(message, "Error", buttons);
                return false;
            }
            else
            {
                return true;
            } //end if
        }

        private static bool CheckIPAddress(string x)
        {
            //Check external IP Address is valid
            IPAddress address;

            try
            {
                address = IPAddress.Parse(x);
            }

            catch (ArgumentNullException)
            {
                //IP Address is invalid
                string message = "Invalid IP Address entered\n\r ";
                MessageBoxButtons buttons = MessageBoxButtons.OK;
                MessageBox.Show(message, "Error", buttons);
                return false;
            }

            catch (FormatException)
            {
                //IP Address is invalid
                string message = "Invalid IP Address entered\n\r ";
                MessageBoxButtons buttons = MessageBoxButtons.OK;
                MessageBox.Show(message, "Error", buttons);
                return false;
            }

            return true;
        }



        // The four threads to monitor for connection requests
        private void iNETTask(string address, int port, int destination)
        {

            Debug.WriteLine("Task 1 started");

            // address is the IP Address, port is the port number
            // destination is 0 for internal -> external. and 1 for external -> internal

            IPEndPoint iEndPointNET = new IPEndPoint(IPAddress.Parse(address), port);

            //Bind and listen
            socketArray[0].Bind(iEndPointNET);
            socketArray[0].Listen(5);

            //Create socket
            Socket internalClientSocketNET = default(Socket);

            bool acceptConnection = true;

            while (acceptConnection)
            {
                try
                {
                    internalClientSocketNET = socketArray[0].Accept();     //accept a connection
                    Thread internalThreadNET = new Thread(new ThreadStart(() => NETConnectionTask(internalClientSocketNET, destination)));
                    internalThreadNET.Name = "Internal NET Connection";
                    internalThreadNET.Start();                         //start the new thread
                }

                catch (SocketException e)
                {
                    // Deal with errors
                    socketArray[0].Close();
                    acceptConnection = false;
                    if (e.Source != null)
                        Debug.WriteLine("IOException source: {0}", e.Source);
                }
                
            } //End While

            Debug.WriteLine("Task 1 ending");

        } // End Task 1

        private void iNxTask(string address, int port, int destination)
        {

            Debug.WriteLine("Task 2 started");

            // address is the IP Address, port is the port number
            // destination is 0 for internal -> external. and 1 for external -> internal

            IPEndPoint iEndPointNET = new IPEndPoint(IPAddress.Parse(address), port);

            //Bind and listen
            socketArray[1].Bind(iEndPointNET);
            socketArray[1].Listen(5);

            //Create socket
            Socket internalClientSocketNET = default(Socket);

            bool acceptConnection = true;

            while (acceptConnection)
            {
                try
                {
                    internalClientSocketNET = socketArray[1].Accept();     //accept a connection
                    Thread internalThreadNET = new Thread(new ThreadStart(() => NxConnectionTask(internalClientSocketNET, destination)));
                    internalThreadNET.Name = "Internal Nx Connection";
                    internalThreadNET.Start();                         //start the new thread
                }

                catch (SocketException e)
                {
                    // Deal with errors
                    socketArray[0].Close();
                    acceptConnection = false;
                    if (e.Source != null)
                        Debug.WriteLine("IOException source: {0}", e.Source);
                }
                
            }//end while

            Debug.WriteLine("Task 2 ending");

        } // End Task 2

        private void eNETTask(string address, int port, int destination)
        {

            Debug.WriteLine("Task 3 started");

            // address is the IP Address, port is the port number
            // destination is 0 for internal -> external. and 1 for external -> internal

            IPEndPoint iEndPointNET = new IPEndPoint(IPAddress.Parse(address), port);

            //Bind and listen
            socketArray[2].Bind(iEndPointNET);
            socketArray[2].Listen(5);

            //Create socket
            Socket internalClientSocketNET = default(Socket);

            bool acceptConnection = true;

            while (acceptConnection)
            {
                try
                {
                    internalClientSocketNET = socketArray[2].Accept();     //accept a connection
                    Thread internalThreadNET = new Thread(new ThreadStart(() => NETConnectionTask(internalClientSocketNET, destination)));
                    internalThreadNET.Name = "External NET Connection";
                    internalThreadNET.Start();                         //start the new thread
                }

                catch (SocketException e)
                {
                    // Deal with errors
                    socketArray[0].Close();
                    acceptConnection = false;
                    if (e.Source != null)
                        Debug.WriteLine("IOException source: {0}", e.Source);
                }

            }//end while

            Debug.WriteLine("Task 3 ending");

        } // End Task 3

        private void eNxTask(string address, int port, int destination)
        {

            Debug.WriteLine("Task 4 started");

            // address is the IP Address, port is the port number
            // destination is 0 for internal -> external. and 1 for external -> internal
                        
            IPEndPoint iEndPointNET = new IPEndPoint(IPAddress.Parse(address), port);

            //Bind and listen
            socketArray[3].Bind(iEndPointNET);
            socketArray[3].Listen(5);

            //Create socket
            Socket internalClientSocketNET = default(Socket);

            bool acceptConnection = true;

            while (acceptConnection)
            {
                try
                {
                    internalClientSocketNET = socketArray[3].Accept();     //accept a connection
                    Thread internalThreadNET = new Thread(new ThreadStart(() => NxConnectionTask(internalClientSocketNET, destination)));
                    internalThreadNET.Name = "External Nx Connection";
                    internalThreadNET.Start();                         //start the new thread
                }

                catch (SocketException e)
                {
                    // Deal with errors
                    socketArray[0].Close();
                    acceptConnection = false;
                    if (e.Source != null)
                        Debug.WriteLine("IOException source: {0}", e.Source);
                }
                
            }//end while

            Debug.WriteLine("Task 4 ending");

        } // End Task 4






        // Two tasks to handle the actual connections
        private void NETConnectionTask(Socket client, int dest)
        {
            //Handle a NETI/NETO

            // dest is 0 for internal -> external. and 1 for external -> internal
            Debug.WriteLine("Do a NET connection");

            //Assume an internal to external
            //Get the last octet of the external IP Address
            int routerClientStn = 0;
            string address = txtExternalIPAddress.Text;
            int lastDot = address.LastIndexOf(".");
            string lastOctet = address.Remove(0,lastDot + 1);
            routerClientStn = int.Parse(lastOctet);

            if (dest > 0)
            {
                //External to Internal connection request
                routerClientStn = int.Parse(txtinternalStnno.Text);
            }
            
                        
            //Un-comment for testing
            //int routerClientStn = int.Parse(testExt);

            byte[] msg = new byte[1100];        //array of 1100 bytes, should be big enough for any QL packet

            //Get the length of the block
            int size = 0;
            int packetLen = 0;
            try
            {
                while (size < 1)       //wait for at least 2 bytes
                {
                   size = client.Receive(msg, 2, SocketFlags.Peek);
                } //end while
                packetLen = msg[1] + 256 * msg[0];

                //Read a data packet from the inital station
                size = client.Receive(msg, packetLen, SocketFlags.None);    //  ,INPUT msg : size=LEN(msg)
            }

            catch (SocketException e)
            {
                //Error while trying to get a data packet
                //Close the client connection and exit
                client.Shutdown(SocketShutdown.Both);
                client.Close();
                if (e.Source != null)
                    Debug.WriteLine("Could not get an inital routing packet from client for a NET connection {0}", e.Source);
                return;
            }

            int routerServerStn = msg[2];   //get the destination station from packet
            int initalSendStn = msg[3];     //get the sending station from packet
            int finalDestStn = msg[8];      //get final destination station number from packet, or zero

            //Set things up for the correct connection request
            string newIP = "127.0.0.1";
            int newPort = 0;
            if (dest > 0)
            {
                //External to Internal network
                newIP = "127.0.0.1";
                newPort = internalPortBaseNET + finalDestStn;
            }
            else
            {
                //Internal to External

                //strip the last octet from my IP Address
                newIP = txtExternalIPAddress.Text;
                char[] MyChar = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' };
                newIP = newIP.TrimEnd(MyChar);
                newIP = newIP + finalDestStn.ToString();     //patch new octet on end
                newPort = externalPortBaseNET;

                //Un-comment for testing
                //newIP = "127.0.0.1";
                //newPort = internalPortBaseNET + finalDestStn;

            } //End if

            //Patch the data packet ready for a successful connection
            msg[0] = 0;
            msg[1] = 14;        // 14 byte packet
            msg[2] = (byte)initalSendStn;     //new destination
            msg[3] = (byte)routerServerStn;   //new sender
            msg[4] = 0;         //block numbers
            msg[5] = 0;
            msg[6] = 1;         //EOF
            msg[7] = 4;         //4 bytes in packet
            msg[8] = 0;
            msg[9] = 0;
            msg[10] = 0; 
            msg[11] = 0;
            msg[12] = 0;
            msg[13] = 0;
            msg[14] = 0;
            msg[15] = 0;

            //Make a connection to the final destination station
           
            Socket ClientSocket = new Socket(AddressFamily.InterNetwork,
                SocketType.Stream, ProtocolType.Tcp);
            IPEndPoint ep2 = new IPEndPoint(IPAddress.Parse(newIP), newPort);

            try
            {
                ClientSocket.Connect(ep2);
            }
            catch (SocketException e)
            {
                //Error while trying make a connection
                //Close the client connection and exit
                client.Shutdown(SocketShutdown.Both);
                client.Close();
                if (e.Source != null)
                    Debug.WriteLine("Could not connect to a server for a NET connection {0}", e.Source);
                return;
            }

            try
            {
                //Send the altered packet back to the client to acknowledge the open on the destination server
                size = client.Send(msg, 0, 14, SocketFlags.None);
            }

            catch (SocketException e)
            {
                //Error while trying to send a data packet
                //Close the client and server connection and exit
                ClientSocket.Shutdown(SocketShutdown.Both);
                ClientSocket.Close();
                client.Shutdown(SocketShutdown.Both);
                client.Close();
                if (e.Source != null)
                    Debug.WriteLine("Could not send inital routing packet response for a NET connection {0}", e.Source);
                return;
            }

            //Start the main loop
            int loop = 0;
            while (loop == 0)
            {
                try
                {
                    //Get the length of the block
                    size = 0;
                    while (size < 1)       //wait for at least 2 bytes
                    {
                        size = client.Receive(msg, 2, SocketFlags.Peek);
                    }
                    packetLen = msg[1] + 256 * msg[0];

                    //Read a data packet from inital station
                    size = client.Receive(msg, packetLen, SocketFlags.None);

                    loop = msg[6];                  //if last packet, exit the loop next time round

                    msg[2] = (byte)finalDestStn;    //patch sender and receiver in packet
                    msg[3] = (byte)routerClientStn;

                    //And send in along to final station
                    size = ClientSocket.Send(msg, 0, packetLen, SocketFlags.None);
                }
                catch (SocketException e)
                {
                    //Error while trying to process a data packet
                    //Close the client and server connection and exit
                    ClientSocket.Shutdown(SocketShutdown.Both);
                    ClientSocket.Close();
                    client.Shutdown(SocketShutdown.Both);
                    client.Close();
                    if (e.Source != null)
                        Debug.WriteLine("Could not process a data packet for a NET connection {0}", e.Source);
                    return;
                }

            } //End while

            // All done close the connection
            ClientSocket.Shutdown(SocketShutdown.Both);
            ClientSocket.Close();
            client.Shutdown(SocketShutdown.Both);
            client.Close();

        } //End task


        private void NxConnectionTask(Socket client, int dest)
        {
            //Handle an FSERVE connection
            //This is more complicated than a NET connection, as it has to work in both directions
            
            // dest is 0 for internal -> external. and 1 for external -> internal
            Debug.WriteLine("Do a Nx connection");

            //Assume an internal to external
            //Get the last octet of the external IP Address
            string address = txtExternalIPAddress.Text;
            int lastDot = address.LastIndexOf(".");
            string lastOctet = address.Remove(0, lastDot + 1);
            int routerClientStn = int.Parse(lastOctet);

            if (dest > 0)
            {
                //External to Internal connection request
                routerClientStn = int.Parse(txtinternalStnno.Text);
            }

            // Un-comment for testing
            //int routerClientStn = int.Parse(testExt);

            byte[] msg = new byte[1100];        //array of 1100 bytes, should be big enough for any QL packet
            
            //Read the open request from the inital station
            //Get the length of the block
            int size = 0;
            int packetLen = 0;

            try
            {
                while (size < 1)       //wait for at least 2 bytes
                {
                    size = client.Receive(msg, 2, SocketFlags.Peek);
                }
                packetLen = msg[1] + 256 * msg[0];

                //Read a data packet
                size = client.Receive(msg, packetLen, SocketFlags.None);    //  ,INPUT msg : size=LEN(msg)
            }

            catch (SocketException e)
            {
                //Error while trying to get a data packet
                //Close the client connection and exit
                client.Shutdown(SocketShutdown.Both);
                client.Close();
                if (e.Source != null)
                    Debug.WriteLine("Could not get inital OPEN packet from client for a Nx connection {0}", e.Source);
                return;
            }

            int routerServerStn = msg[2];   //get the destination station from packet
            int initalSendStn = msg[3];     //get the sending station from packet

            //get final destination station from filename/device name
            int nameLen = (256 * msg[14]) + msg[15];    //length of file/device name in OPEN

            int finalDestStn = 0;
            bool loop = true;

            int ptr = 17;           //pointer in string (start of destination station number)
            int x = 0;              //number of digits in station number

            if ((msg[16] == 78) || (msg[16] == 110))    //OR
            {
                //The string starts with a 'n' or a 'N'
                while (loop)
                {
                    if ((msg[ptr] >= 48) && (msg[ptr] <= 57))   //AND
                    {
                        //it's a number
                        x = x + 1;          //one more digit
                        ptr = ptr + 1;      //move pointer
                        continue;   //NEXT loop
                    } //end if

                    if (msg[ptr] == 95)    //underscore
                    {
                        loop = false;
                        if (x >= 4)
                        {
                            x = 0;          //too many characters
                        } // end if
                        break;  //EXIT loop
                    } //end if

                    //Unrecognized character found
                    x = 0;
                    loop = false;
                    
                } // end while
                
            } //end if

            //On good exit of loop, ptr is the underscore, x is number of digits
            //On bad exit, x is zero

            if (x == 0)
            {
                //Somethings wrong, send a 'bad name' error back to the client
                //Patch the data packet ready for error
                msg[0] = 0;
                msg[1] = 14;        // 14 byte packet
                msg[2] = (byte)initalSendStn;       //new destination
                msg[3] = (byte)routerServerStn;     //new sender
                msg[4] = 0;         //block numbers
                msg[5] = 0;
                msg[6] = 1;         //EOF
                msg[7] = 4;         //4 bytes in packet
                msg[8] = 0;
                msg[9] = 0;
                msg[10] = 255;      //error number -12 Bad Name
                msg[11] = 255;
                msg[12] = 255;
                msg[13] = 244;
                msg[14] = 0;
                msg[15] = 0;

                try
                {
                    //Send the packet back to the client to acknowledge the error
                    size = client.Send(msg, 0, 14, SocketFlags.None);
                }

                catch (SocketException e)
                {
                    //Error while trying to send a packet
                    //Close the client connection and exit
                    client.Shutdown(SocketShutdown.Both);
                    client.Close();
                    if (e.Source != null)
                        Debug.WriteLine("Could not send error packet to client for a Nx connection {0}", e.Source);
                    return;
                }

                //Close the client, and end the job
                client.Shutdown(SocketShutdown.Both);
                client.Close();
                return;

            } //endif

            //Extract the final destination station number
            int multiplier = 1;
            int index = ptr - 1;
            int y = x;

            while (y > 0)
            {
                finalDestStn = finalDestStn + ((msg[index] - 48) * multiplier);
                multiplier = multiplier * 10;
                index = index - 1;
                y = y - 1;
            } //end while

            //Patch the data packet to remove the final destination 
            nameLen = nameLen - x - 2;          //number of digits minus the 'n' and '_'
            //Alter string length
            msg[14] = Convert.ToByte(nameLen / 256);        // DIV 256
            msg[15] = Convert.ToByte(nameLen % 256);        // MOD 256

            //Shuffle name up over the final destination station
            for (y = 0; y <= nameLen; y++)
            {
                msg[16 + y] = msg[ptr + 1 + y];
            } //end for
            
            //Set things up for the correct connection request to destination FSERVE
            string newIP = "127.0.0.1";
            int newPort = 0;
            if (dest > 0)
            {
                //External to Internal network
                newIP = "127.0.0.1";
                newPort = internalPortBaseNx + finalDestStn;
            }
            else
            {
                //Internal to External

                //strip the last octet from my IP Address
                newIP = txtExternalIPAddress.Text;
                char[] MyChar = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' };
                newIP = newIP.TrimEnd(MyChar);
                newIP = newIP + finalDestStn.ToString();     //patch new octet on end
                newPort = externalPortBaseNx;

                //Un-comment for testing
                //newIP = "127.0.0.1";
                //newPort = internalPortBaseNx + finalDestStn;

            } //End if


            //Patch packet for onward transmission
            msg[2] = (byte)finalDestStn;   //new destination
            msg[8] = msg[3];               //copy sender for NAS_SHARE checks

            msg[3] = (byte)routerClientStn;   //new sender

            nameLen = nameLen + 6;      //plus the word length & open type
            msg[7] = Convert.ToByte(nameLen % 256);        //set new data length
            nameLen = nameLen + 10;     //plus header length

            msg[0] = Convert.ToByte(nameLen / 256);        // DIV 256
            msg[1] = Convert.ToByte(nameLen % 256);        // MOD 256  set new packet length


            //Make a connection to destination FSERVE 
            Socket ClientSocket = new Socket(AddressFamily.InterNetwork,
                SocketType.Stream, ProtocolType.Tcp);
            IPEndPoint ep2 = new IPEndPoint(IPAddress.Parse(newIP), newPort);
            try
            {
                ClientSocket.Connect(ep2);
            }

            catch (SocketException e)
            {
                //Error while trying make a connection
                //Close the client connection and exit
                client.Shutdown(SocketShutdown.Both);
                client.Close();
                if (e.Source != null)
                    Debug.WriteLine("Could not connect to a server for a Nx connection {0}", e.Source);
                return;
            }

            try
            {
                //Send the packet on to the destination server
                size = ClientSocket.Send(msg, 0, nameLen, SocketFlags.None);
            }

            catch (SocketException e)
            {
                //Error while trying to forward OPEN packet
                //Close the client and server connection and exit
                ClientSocket.Shutdown(SocketShutdown.Both);
                ClientSocket.Close();
                client.Shutdown(SocketShutdown.Both);
                client.Close();
                if (e.Source != null)
                    Debug.WriteLine("Could not forward OPEN packet for a Nx connection {0}", e.Source);
                return;
            }
            
            byte blockType = msg[6];        //get the value of D0 for the TRAP #2 call
                                            //so we know if it's a DELETE request


            //Receive the reply from FSERVE
            //Get the length of the block
            size = 0;

            try
            {
                while (size < 1)       //wait for at least 2 bytes
                {
                    size = ClientSocket.Receive(msg, 2, SocketFlags.Peek);
                }
                packetLen = msg[1] + 256 * msg[0];

                //Read a data packet
                size = ClientSocket.Receive(msg, packetLen, SocketFlags.None);

                //Patch the data packet
                msg[2] = (byte)initalSendStn;     //new destinition
                msg[3] = (byte)routerServerStn;   //new sender

                //Forward the packet to the inital client station
                size = client.Send(msg, 0, packetLen, SocketFlags.None);
            }

            catch (SocketException e)
            {
                //Error while trying to return result of OPEN
                //Close the client and server connection and exit
                ClientSocket.Shutdown(SocketShutdown.Both);
                ClientSocket.Close();
                client.Shutdown(SocketShutdown.Both);
                client.Close();
                if (e.Source != null)
                    Debug.WriteLine("Could not process OPEN result for a Nx connection {0}", e.Source);
                return;
            }

            //If there was a delete, or an open that failed, then terminate the job
            if ((blockType == 4) || (msg[13] > 0))        // OR
            {
                //IO_DELETE or error, close everything and end job
                ClientSocket.Shutdown(SocketShutdown.Both);
                ClientSocket.Close();
                client.Shutdown(SocketShutdown.Both);
                client.Close();
                return;
            } //end if


            //The client and the server sides are now open and we can start the main I/O loop
            //looking for a close packet
            loop = true;

            while (loop)
            {
                try
                {
                    // Read a packet fron client station
                    //Get the length of the block
                    size = 0;
                    while (size < 1)       //wait for at least 2 bytes
                    {
                        size = client.Receive(msg, 2, SocketFlags.Peek);
                    }
                    packetLen = msg[1] + 256 * msg[0];

                    //Read a data packet
                    size = client.Receive(msg, packetLen, SocketFlags.None);    //  ,INPUT msg : size=LEN(msg)

                    blockType = msg[6];          //get the value of D0 for the TRAP call

                    //Patch the data packet
                    msg[2] = (byte)finalDestStn;        //new destinition
                    msg[3] = (byte)routerClientStn;     //new sender

                    //Send the packet on to the destination FSERVE
                    size = ClientSocket.Send(msg, 0, packetLen, SocketFlags.None);

                    //Receive the reply from FSERVE
                    //Get the length of the block
                    size = 0;
                    while (size < 1)       //wait for at least 2 bytes
                    {
                        size = ClientSocket.Receive(msg, 2, SocketFlags.Peek);
                    }
                    packetLen = msg[1] + 256 * msg[0];

                    //Read a data packet
                    size = ClientSocket.Receive(msg, packetLen, SocketFlags.None);

                    //Patch the data packet
                    msg[2] = (byte)initalSendStn;     //new destinition
                    msg[3] = (byte)routerServerStn;   //new sender

                    //Forward the packet to the inital client
                    size = client.Send(msg, 0, packetLen, SocketFlags.None);
                } //end try

                catch (SocketException e)
                {
                    //Error while trying to process a data packet
                    //Close the client and server connection and exit
                    ClientSocket.Shutdown(SocketShutdown.Both);
                    ClientSocket.Close();
                    client.Shutdown(SocketShutdown.Both);
                    client.Close();
                    if (e.Source != null)
                        Debug.WriteLine("Could not process a data packet for a Nx connection {0}", e.Source);
                    return;
                }

                if (blockType == 2)
                {
                    //IO_CLOSE, exit loop
                    loop = false;
                } //end if
                
            } //end while

            // All done close the connection
            ClientSocket.Shutdown(SocketShutdown.Both);
            ClientSocket.Close();
            client.Shutdown(SocketShutdown.Both);
            client.Close();

        } //End task


        private void btnAbout_Click(object sender, EventArgs e)
        {
            
            string message = "IP Router Version " + programVersion + "\n\r Martin Head   2019";
            MessageBoxButtons buttons = MessageBoxButtons.OK;
            MessageBox.Show(message, "About", buttons);
        }


        private void btnStop_Click(object sender, EventArgs e)
        {
            txtinternalStnno.Enabled = true;
            txtExternalIPAddress.Enabled = true;
            btnStart.Enabled = true;
            btnStart.Visible = true;
            btnStop.Enabled = false;
            btnStop.Visible = false;

            //Stop the four main accept jobs running by closing their 
            //sockets, Producing an error.
            Debug.WriteLine("Attempting to close main tasks");
            socketArray[0].Close();
            socketArray[1].Close();
            socketArray[2].Close();
            socketArray[3].Close();
            Debug.WriteLine("Tasks closed?");


        }

        
    } //End of Form1
}
